# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from io import BytesIO

from mo_files import File
from mo_imports import expect, export
from mo_logs import logger

from mo_json import JxType, JX_TEXT
from mo_streams._utils import chunk_bytes, Stream

ObjectStream, StringStream, File_usingStream, Typer = expect(
    "ObjectStream", "StringStream", "File_usingStream", "Typer"
)


DEBUG = True


class ByteStream(Stream):
    def __init__(self, reader):
        self.verbose = DEBUG
        self.reader: BytesIO = reader

    def close(self):
        self.reader.close()

    def from_zip(self) -> ObjectStream:
        """
        RETURN A STREAM OF Files
        """
        from zipfile import ZipFile

        def read():
            # ZIP HAS DIRACTORY AT END OF FILE, MUST READ WHOLE THING
            reader = self.reader
            try:
                if not reader.seekable():
                    reader = BytesIO(b"".join(chunk_bytes(reader)))
                with ZipFile(reader, mode="r") as archive:
                    for info in archive.filelist:
                        yield File_usingStream(
                            info.filename,
                            lambda: ByteStream(archive.open(info.filename, "r")),
                        ), {"name": info.filename}
            finally:
                reader.close()

        return ObjectStream(read(), Typer(type_=File_usingStream), JxType(name=JX_TEXT))

    def from_zst(self):
        """
        :param stream:
        :return:
        """
        from zstandard import ZstdDecompressor

        stream_reader = ZstdDecompressor().stream_reader(self.reader, closefd=True)
        return ByteStream(stream_reader)

    def from_tar(self):
        """
        return a stream of files
        """
        import tarfile

        tf = tarfile.open(mode="r:", fileobj=self.reader)

        def file(info):
            reader = tf.extractfile(info)
            if reader is None:
                # directories
                return File_usingStream(info.name, lambda: None)
            else:
                return File_usingStream(
                    info.name, lambda: ByteStream(tf.extractfile(info))
                )

        def read():
            try:
                while True:
                    info = tf.next()
                    if not info:
                        return
                    yield file(info), {"name": info.name}
            finally:
                self.reader.close()

        return ObjectStream(read(), Typer(type_=File_usingStream), JxType(name=JX_TEXT))

    def to_zst(self):
        from zstandard import ZstdCompressor

        return ByteStream(ZstdCompressor().stream_reader(self.reader))

    def utf8(self):
        def read():
            for data in chunk_bytes(self.reader):
                yield data.decode("utf8")
            self.reader.close()

        return StringStream(read())

    def lines(self):
        return self.utf8.lines

    def chunk(self, size=8192):
        return ObjectStream(chunk_bytes(self.reader, size), b"", bytes, {}, JxType())

    def write(self, file):
        file = File(file)
        with open(file.os_path, "wb") as f:
            for d in chunk_bytes(self.reader):
                f.write(d)

    def to_bytes(self):
        return b"".join(chunk_bytes(self.reader))

    def to_s3(self, *, name, bucket):
        import boto3

        s3_client = boto3.client("s3")
        try:
            return s3_client.upload_fileobj(self.reader, bucket, name)
        except Exception as cause:
            if self.verbose:
                logger.warn("problem with s3 upload", cause=cause)


export("mo_streams._utils", ByteStream)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from mo_dots import Data
from mo_dots.lists import is_many, is_finite
from mo_files import File
from mo_imports import export

from mo_future import first
from mo_json import JxType, JX_TEXT
from mo_streams._utils import Stream, Reader
from mo_streams.byte_stream import ByteStream
from mo_streams.empty_stream import EmptyStream
from mo_streams.files import content, File_usingStream
from mo_streams.function_factory import it
from mo_streams.object_stream import ObjectStream, ERROR, WARNING, NONE
from mo_streams.string_stream import StringStream
from mo_streams.type_utils import Typer, CallableTyper


def stream(value):
    if isinstance(value, (dict, Data)):
        kv = first(value.items())
        if not kv:
            return EmptyStream()
        _, example = kv
        return ObjectStream(
            ((v, {"key": k}) for k, v in value.items()),
            Typer(example=example),
            JxType(key=JX_TEXT),
        )
    elif isinstance(value, bytes):
        return ByteStream(Reader(iter([value])))
    elif isinstance(value, str):
        return StringStream(iter([value]))
    elif value == None:
        return EmptyStream
    elif isinstance(value, Stream):
        return value
    elif isinstance(value, type(range(1))):
        return ObjectStream(
            ((v, {}) for v in value), Typer(example=value.stop), JxType()
        )
    elif is_finite(value):
        example = first(value)
        def read_from_list():
            for v in value:
                yield v, {}

        return ObjectStream(read_from_list(), Typer(example=example), JxType())
    elif is_many(value):
        example = first(value)
        def read():
            yield example, {}
            for v in value:
                yield v, {}

        return ObjectStream(read(), Typer(example=example), JxType())
    else:
        return ObjectStream(iter([(value, {})]), Typer(example=value), JxType())


ANNOTATIONS = {
    (str, "encode"): CallableTyper(type_=bytes),
    (File_usingStream, "content"): CallableTyper(type_=ByteStream),
    (File, "content"): CallableTyper(type_=ByteStream),
    (ByteStream, "utf8"): CallableTyper(type_=StringStream),
}

export("mo_streams.object_stream", stream)
export("mo_streams.type_utils", ANNOTATIONS)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from typing import Any, Iterator, Dict, Tuple
from zipfile import ZIP_STORED

from mo_future import zip_longest, first
from mo_imports import expect, export
from mo_json import JxType, JX_INTEGER
from mo_logs import logger

from mo_files import File
from mo_streams import ByteStream
from mo_streams._utils import (
    Reader,
    Writer,
    chunk_bytes,
    Stream,
)
from mo_streams.byte_stream import DEBUG
from mo_streams.files import File_usingStream
from mo_streams.function_factory import factory
from mo_streams.type_utils import Typer, LazyTyper

stream = expect("stream")

ERROR = {}
WARNING = {}
NONE = {}


class ObjectStream(Stream):
    """
    A STREAM OF OBJECTS
    """

    def __init__(self, values, datatype, schema):
        if not isinstance(datatype, Typer) or isinstance(datatype, LazyTyper):
            logger.error(
                "expecting datatype to be Typer not {{type}}",
                type=datatype.__class__.__name__,
            )
        self._iter: Iterator[Tuple[Any, Dict[str, Any]]] = values
        self.type_: Typer = datatype
        self._schema = schema

    def __getattr__(self, item):
        type_ = getattr(self.type_, item)

        def read():
            for v, a in self._iter:
                try:
                    yield getattr(v, item), a
                except (StopIteration, GeneratorExit):
                    raise
                except Exception as cause:
                    DEBUG and logger.warn(
                        "can not get attribute {{item|quote}}", cause=cause
                    )
                    yield None, a

        return ObjectStream(read(), type_, self._schema)

    def __call__(self, *args, **kwargs):
        type_ = self.type_(*args, **kwargs)

        if type_.type_ == bytes:

            def read_bytes():
                for m, a in self._iter:
                    try:
                        yield m(*args, **kwargs)
                    except (StopIteration, GeneratorExit):
                        raise
                    except Exception:
                        yield None

            return ByteStream(Reader(read_bytes()))

        def read():
            for m, a in self._iter:
                try:
                    yield m(*args, **kwargs), a
                except (StopIteration, GeneratorExit):
                    raise
                except Exception:
                    yield None, a

        return ObjectStream(read(), type_, self._schema)

    def map(self, accessor):
        if isinstance(accessor, str):
            type_ = getattr(self.type_, accessor)
            return ObjectStream(
                ((getattr(v, accessor), a) for v, a in self._iter), type_, self._schema
            )
        fact = factory(accessor, self.type_)
        do_accessor = fact.build(self.type_, self._schema)

        def read():
            for v, a in self._iter:
                try:
                    yield do_accessor(v, a), a
                except (StopIteration, GeneratorExit):
                    raise
                except Exception as cause:
                    yield None, a

        if isinstance(fact.type_, LazyTyper):
            type_ = fact.type_._resolver(self.type_)
        else:
            type_ = fact.type_

        return ObjectStream(read(), type_, self._schema)

    def filter(self, predicate):
        fact = factory(predicate, self.type_)
        filteror = fact.build(self.type_, self._schema)

        def read():
            for v, a in self._iter:
                try:
                    if filteror(v, a):
                        yield v, a
                except (StopIteration, GeneratorExit):
                    raise
                except Exception as cause:
                    pass

        return ObjectStream(read(), self.type_, self._schema)


    def attach(self, **kwargs):
        facts = {k: factory(v) for k, v in kwargs.items()}

        more_schema = JxType()  # NOT AT REAL TYPE, WE ADD PYTHON TYPES ON THE LEAVES
        for k, f in facts.items():
            setattr(more_schema, k, f.type_)

        mapper = {k: f.build(f.type_, self._schema) for k, f in facts.items()}

        def read():
            for v, a in self._iter:
                yield v, {**a, **{k: m(v, a) for k, m in mapper.items()}}

        return ObjectStream(read(), self.type_, self._schema | more_schema)

    def exists(self):
        def read():
            for v, a in self._iter:
                if v != None:
                    yield v, a

        return ObjectStream(read(), self.type_, self._schema)

    def enumerate(self):
        def read():
            for i, (v, a) in enumerate(self._iter):
                yield v, {**a, "index": i}

        return ObjectStream(read(), self.type_, self._schema | JxType(index=JX_INTEGER))

    def flatten(self):
        def read():
            for v, a in self._iter:
                for vv, aa in stream(v)._iter:
                    yield vv, {**a, **aa}

        return ObjectStream(read(), self.type_, self._schema)

    def reverse(self):
        def read():
            yield from reversed(list(self._iter))

        return ObjectStream(read(), self.type_, schema=self._schema)

    def sort(self, *, key=None, reverse=0):
        if key:
            key = lambda t: key(t[0])

        def read():
            yield from sorted(self._iter, key=key, reverse=reverse)

        return ObjectStream(read(), self.type_, self._schema)

    def distinct(self):
        def read():
            acc = set()
            for v, a in self._iter:
                if v in acc:
                    continue
                acc.add(v)
                yield v, a

        return ObjectStream(read(), self.type_, self._schema)

    def append(self, value):
        def read():
            yield from self._iter
            yield value, {}

        return ObjectStream(read(), self.type_, self._schema)

    def extend(self, values):
        suffix = stream(values)

        def read():
            yield from self._iter
            yield from suffix._iter

        return ObjectStream(read(), self.type_, self._schema | suffix._schema)

    def zip(self, *others):
        streams = [stream(o) for o in others]

        def read():
            yield from zip_longest(self._iter, *(s._iter for s in streams))

        return TupleStream(
            read(),
            self._example,
            self.type_,
            sum((s._schema for s in streams), JxType()),
        )

    def limit(self, count):
        def read():
            try:
                for i in range(count):
                    yield next(self._iter)
            except StopIteration:
                pass

        return ObjectStream(read(), self.type_, self._schema)

    def materialize(self):
        return ObjectStream(list(self._iter), self.type_, self._schema)

    def to_list(self):
        return list(v for v, _ in self._iter)

    def count(self):
        return sum(1 for _ in self._iter)

    def to_dict(self, key=None):
        """
        CONVERT STREAM TO dict
        :param key: CHOOSE WHICH ANNOTATION IS THE KEY
        """
        if key is None:
            candidates = self._schema.__dict__.keys()
            if len(candidates) != 1:
                logger.error(
                    "expecting attachment to have just one property, not {{num}}",
                    num=len(candidates),
                )
            key = first(candidates)

        return {a[key]: v for v, a in self._iter}

    def to_zip(
        self, compression=ZIP_STORED, allowZip64=True, compresslevel=None,
    ):
        from zipfile import ZipFile, ZipInfo

        type_ = self.type_.type_
        if type_ is File:
            pass
        elif type_ is File_usingStream:
            pass
        else:
            raise NotImplementedError("expecting stream of Files")

        def read():
            mode = "w"
            writer = Writer()
            with ZipFile(
                writer,
                mode=mode,
                compression=compression,
                allowZip64=allowZip64,
                compresslevel=compresslevel,
            ) as archive:
                for file, _ in self._iter:
                    info = ZipInfo(file.rel_path)
                    with archive.open(info, mode=mode) as target:
                        for chunk in chunk_bytes(file.bytes()):
                            target.write(chunk)
                            yield writer.read()

            yield writer.read()
            writer.close()

        return ByteStream(Reader(read()))


export("mo_streams.byte_stream", ObjectStream)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import inspect
import os
from io import RawIOBase, BytesIO
from typing import BinaryIO

from mo_dots.lists import Log
from mo_imports import expect
from mo_logs import logger

ByteStream = expect("ByteStream")
START, CURRENT, END = 0, 1, 2


class Stream:
    pass


class Reader(BinaryIO):
    """
    WRAP A GENERATOR WITH A FILE-LIKE OBJECT
    """

    def __init__(self, chunks):
        self._chunks = chunks
        self.residue = b""
        self.count = 0

    def readable(self):
        return True

    def read(self, size=-1):
        if not self._chunks:
            return self._more(size)

        try:
            if size == -1:
                data = next(self._chunks)
                self.count += len(data)
                return data

            while len(self.residue) < size:
                chunk = next(self._chunks)
                self.residue += chunk
        except StopIteration:
            self._chunks = None
        return self._more(size)

    def _more(self, size):
        data = self.residue[:size]
        self.residue = self.residue[size:]
        self.count += len(data)
        return data

    def tell(self):
        return self.count

    def seek(self, position, whence=START):
        if whence == END:
            everything = BytesIO(b"".join(self._chunks))

        if self.count > position:
            raise NotImplementedError()
        self.read(position - self.count)


class Writer(RawIOBase):
    """
    REPLACE IO SO THAT WE CAN read() THE RESULTING
    """

    def __init__(self):
        self._buffer = b""

    def writable(self):
        return True

    def seekable(self):
        return False

    def write(self, b):
        if self.closed:
            raise Exception("stream was closed")
        self._buffer += b
        return len(b)

    def read(self, size=-1):
        if size == -1 or size > len(self._buffer):
            chunk = self._buffer
            self._buffer = b""
            return chunk

        chunk = self._buffer[:size]
        self._buffer = self._buffer[size:]
        return chunk

    def content(self):
        return ByteStream(self)

    def size(self):
        return len(self._buffer)


def chunk_bytes(reader, size=4096):
    """
    WRAP A FILE-LIKE OBJECT TO LOOK LIKE A GENERATOR
    """

    if isinstance(reader, ByteStream):
        reader = reader.reader
    if isinstance(reader, Reader):
        return reader._chunks

    def read():
        """
        :return:
        """
        try:
            while True:
                data = reader.read(size)
                if not data:
                    return
                yield data
        except Exception as e:
            Log.error("Problem iterating through stream", cause=e)
        finally:
            try:
                reader.close()
            except Exception as cause:
                pass

    return read()


def is_function(value):
    if type(value).__name__ == "function":
        return True
    if isinstance(value, type):
        return True
    if hasattr(value, "__call__"):
        logger.error("not expected")
    return False


def arg_spec(type_, item):
    for name, func in inspect.getmembers(type_):
        if name != item:
            continue
        return inspect.getfullargspec(func)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import inspect
from types import FunctionType

from mo_logs import logger, Except
from mo_logs.exceptions import ERROR, get_stacktrace

from mo_streams.type_utils import Typer, LazyTyper, CallableTyper, UnknownTyper

_get = object.__getattribute__
_set = object.__setattr__


class FunctionFactory:
    """
    See mo-streams/docs/function_factory.md
    """

    def __init__(self, builder, type_, desc):
        if not isinstance(type_, Typer):
            logger.error("expecting type, not {{type}}", type=type_)
        _set(self, "build", builder)
        _set(self, "type_", type_)
        _set(self, "_desc", desc)

    def __getattr__(self, item):
        def builder(type_, _schema):
            s = _get(self, "build")(type_, _schema)
            if item in _schema:

                def func(v, a):
                    return a[item]

                return func
            elif isinstance(item, FunctionFactory):
                i = item.build(type_, _schema)
                return lambda v, a: getattr(s(v, a), i(v, a))
            else:
                return lambda v, a: getattr(s(v, a), item)

        return FunctionFactory(
            builder, getattr(_get(self, "type_"), item), f"{self}.{item}"
        )

    def __eq__(self, other):
        func_other = factory(other)

        def builder(type_, _schema):
            s = self.build(type_, _schema)
            o = func_other.build(type_, _schema)

            def func(v, a):
                return s(v, a) == o(v, a)

            return func

        return FunctionFactory(builder, Typer(type_=bool), f"{other} == {self}")

    def __radd__(self, other):
        if isinstance(other, FunctionFactory):
            func_other = other
        else:
            func_other = FunctionFactory(
                lambda v, a: other, Typer(example=other), str(other)
            )

        def builder(type_, _schema):
            s = self.build(type_, _schema)
            o = func_other.build(type_, _schema)

            def func(v, a):
                return o + s(v, a)

            return func

        type_ = Typer(example=other) + _get(self, "type_")
        return FunctionFactory(builder, type_, f"{other} + {self}")

    def __call__(self, *args, **kwargs):
        args = [factory(a) for a in args]
        kwargs = {k: factory(v) for k, v in kwargs.items()}

        def builder(type_, _schema):
            s = self.build(type_, _schema)
            _args = [a.build(type_, _schema) for a in args]
            _kwargs = {k: v.build(type_, _schema) for k, v in kwargs.items()}

            def func(v, a):
                return s(v, a)(
                    *(f(v, a) for f in _args),
                    **{k: f(v, a) for k, f in _kwargs.items()},
                )

            return func

        desc_args = [str(a) for a in args]
        desc_args.extend(f"{k}={v}" for k, v in kwargs.items())
        params = ",".join(desc_args)

        return FunctionFactory(builder, self.type_(), f"{self}({params})")

    def __str__(self):
        return _get(self, "_desc")


def factory(item, self_type=None, return_type=None):
    if isinstance(item, (str, bytes, bool, int, float)):
        # CONSTANT
        def build_constant(type_, _schema):
            return lambda v, a: item

        return FunctionFactory(build_constant, Typer(example=item), f"{item}")
    elif isinstance(item, FunctionFactory):
        return item
    else:
        normalized_func, return_type = wrap_func(item, return_type=return_type)

        def builder3(type_, _schema):
            return normalized_func

        return FunctionFactory(builder3, return_type, f"returning {return_type}")


def build(item):
    if isinstance(item, FunctionFactory):
        return item.build

    def builder(type_, _schema):
        return lambda v, a: item

    return builder


# build list of single arg builtins, that can be used as parse actions
singleArgBuiltins = [
    sum,
    len,
    sorted,
    reversed,
    list,
    tuple,
    set,
    any,
    all,
    min,
    max,
]

singleArgTypes = [
    int,
    float,
    str,
    bool,
    complex,
    dict,
]


def wrap_func(func, return_type=None):
    try:
        func_name = getattr(func, "__name__", getattr(func, "__class__").__name__)
    except Exception:
        func_name = str(func)

    if func in singleArgBuiltins:
        spec = inspect.getfullargspec(func)
    elif func.__class__.__name__ == "staticmethod":
        func = func.__func__
        spec = inspect.getfullargspec(func)
    elif func.__class__.__name__ == "builtin_function_or_method":
        spec = inspect.getfullargspec(func)
    elif func in singleArgTypes:
        spec = inspect.FullArgSpec(["value"], None, None, None, [], None, {})
        return_type = func
    elif isinstance(func, type):
        spec = inspect.getfullargspec(func.__init__)
        new_func = func.__call__
        # USE ONLY FIRST PARAMETER
        num_args = len(spec.args) - 1  # ASSUME self IS FIRST ARG
        if num_args == 0:

            def wrap_init0(val, att):
                return new_func()

            return wrap_init0, Typer(type_=func)
        else:

            def wrap_init1(val, att):
                return new_func(val)

            return wrap_init1, Typer(type_=func)
    elif isinstance(func, FunctionType):
        spec = inspect.getfullargspec(func)
    elif hasattr(func, "__call__"):
        spec = inspect.getfullargspec(func)

    if spec.varargs:
        num_args = 3
    elif spec.args and spec.args[0] in ["cls", "self"]:
        num_args = len(spec.args) - 1
    else:
        num_args = len(spec.args)

    if not return_type:
        return_type = spec.annotations.get("return")
    if return_type:
        return_type = Typer(type_=return_type)
    else:
        cause = Except(
            ERROR,
            "expecting {{function}} to have annotated return type",
            {"function": func_name},
            trace=get_stacktrace(start=3)
        )
        return_type = UnknownTyper(cause)

    if num_args == 0:
        def wrapper0(val, att):
            return func()

        wrapper = wrapper0
    elif num_args == 1:

        def wrapper1(val, att):
            return func(val)

        wrapper = wrapper1
    else:
        return func, return_type

    # copy func name to wrapper for sensible debug output
    wrapper.__name__ = func_name

    return wrapper, return_type


class TopFunctionFactory(FunctionFactory):
    """
    it(x)  RETURNS A FunctionFactory FOR x
    """

    def __call__(self, value):
        if isinstance(value, FunctionFactory):
            logger.error("don't do this")

        def builder(type_, _schema):
            return lambda v, a: value

        if isinstance(value, type):
            return FunctionFactory(builder, CallableTyper(type_=value), f"{value}")

        return FunctionFactory(builder, Typer(type_=type(value)), f"{value}")

    def __str__(self):
        return "it"


def noop(v, a) -> LazyTyper:
    return v


it = factory(noop)
it.__class__ = TopFunctionFactory
